
package com.ichi2.anki.dialogs;

import android.content.DialogInterface;
import android.content.res.Resources;
import android.os.Bundle;
import android.support.v4.app.DialogFragment;
import android.widget.CompoundButton.OnCheckedChangeListener;

import com.ichi2.anki.R;
import com.ichi2.themes.StyledDialog;

public class ExportDialog extends DialogFragment {

    public interface ExportDialogListener {

        public void exportApkg(String path, Long did, boolean includeSched, boolean includeMedia);


        public void dismissAllDialogFragments();
    }

    private final int INCLUDE_SCHED = 0;
    private final int INCLUDE_MEDIA = 1;
    private boolean mIncludeSched = true;
    private boolean mIncludeMedia = false;


    /**
     * A set of dialogs which deal with importing a file
     * 
     * @param dialogType An integer which specifies which of the sub-dialogs to show
     * @param dialogMessage An optional string which can be used to show a custom message or specify import path
     */
    public static ExportDialog newInstance(String dialogMessage, Long did) {
        ExportDialog f = new ExportDialog();
        Bundle args = new Bundle();
        args.putLong("did", did);
        args.putString("dialogMessage", dialogMessage);
        f.setArguments(args);
        return f;
    }


    public static ExportDialog newInstance(String dialogMessage) {
        ExportDialog f = new ExportDialog();
        Bundle args = new Bundle();
        args.putString("dialogMessage", dialogMessage);
        f.setArguments(args);
        return f;
    }


    @Override
    public StyledDialog onCreateDialog(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        Resources res = getResources();
        final Long did = getArguments().getLong("did", -1L);
        mIncludeSched = did == -1L;
        final String[] items = { res.getString(R.string.export_include_schedule),
                res.getString(R.string.export_include_media) };
        final boolean[] checked = { mIncludeSched, mIncludeMedia };
        StyledDialog.Builder builder = new StyledDialog.Builder(getActivity());
        setCancelable(true);
        builder.setMessage(getArguments().getString("dialogMessage"));
        builder.setMultiChoiceItems(items, checked, new DialogInterface.OnClickListener() {
            // indexSelected contains the index of item (of which checkbox checked)
            @Override
            public void onClick(DialogInterface dialog, int indexSelected) {
                switch (indexSelected) {
                    case INCLUDE_SCHED:
                        mIncludeSched = !mIncludeSched;
                        break;
                    case INCLUDE_MEDIA:
                        mIncludeMedia = !mIncludeMedia;
                        break;
                }
            }
        }, (OnCheckedChangeListener) null);

        builder.setPositiveButton(res.getString(R.string.dialog_ok), new DialogInterface.OnClickListener() {

            @Override
            public void onClick(DialogInterface dialog, int which) {
                ((ExportDialogListener) getActivity()).exportApkg(null, did != -1L ? did : null, mIncludeSched,
                        mIncludeMedia);
                dismissAllDialogFragments();
            }

        });
        builder.setNegativeButton(res.getString(R.string.dialog_cancel), null);
        return builder.create();
    }


    public void dismissAllDialogFragments() {
        ((ExportDialogListener) getActivity()).dismissAllDialogFragments();
    }

}
